from typing import TYPE_CHECKING, Union, Optional, Callable, Dict

import ezdxf.entities

from geogeometry.geo_io.readers.dxf_components.DxfEntityHandler import DxfEntityHandler

if TYPE_CHECKING:
    from geogeometry.geometry.triangulation.Triangulation import Triangulation
    from geogeometry.geometry.polyline.Polyline import Polyline

output_types = Optional[Union['Triangulation', 'Polyline']]
handler_method = Callable[..., output_types]


class EntityHandlerRegistry:

    def __init__(self):
        self.handlers: Dict[str, handler_method] = {}
        self.setHandlers()

    def setHandlers(self) -> None:
        self.handlers["POLYLINE"] = DxfEntityHandler.handlePolyline
        self.handlers["LINE"] = DxfEntityHandler.handleLine

    def getHandler(self, entity_type: str) -> handler_method:
        return self.handlers.get(entity_type)



from typing import Optional, TYPE_CHECKING, Literal

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorProperties(object):

    def __init__(self):

        self.n0: Optional[np.ndarray] = None
        self.n1: Optional[np.ndarray] = None

        self.dimensions: Optional[int] = None

        self.length: Optional[float] = None
        self.unitary: Optional[bool] = None

        self.normalized_vector: Optional['Vector'] = None

        self.dip: Optional[float] = None
        self.dipdir: Optional[float] = None

    def setN0(self, n0: np.ndarray) -> None:
        self.n0 = n0

    def setN1(self, n1: np.ndarray) -> None:
        self.n1 = n1

    def setDimensions(self, dimensions: Literal[2, 3]) -> None:
        self.dimensions = dimensions

    def setLength(self, length: float) -> None:
        self.length = length

    def setUnitary(self, unitary: bool) -> None:
        self.unitary = unitary

    def setNormalizedVector(self, normalized_vector: 'Vector') -> None:
        self.normalized_vector = normalized_vector

    def setDip(self, dip: float) -> None:
        self.dip = dip

    def setDipdir(self, dipdir: float) -> None:
        self.dipdir = dipdir

    def getN0(self) -> np.ndarray:
        return self.n0

    def getN1(self) -> np.ndarray:
        return self.n1

    def getNodes(self) -> np.ndarray:
        return np.array([self.n0, self.n1])

    def getLength(self) -> float:
        return self.length

    def isUnitary(self) -> bool:
        return self.unitary

    def getNormalizedVector(self) -> 'Vector':
        return self.normalized_vector

    def getDip(self) -> float:
        return self.dip

    def getDipdir(self) -> float:
        return self.dipdir



import os
import unittest
from geogeometry import CadReader


class TestReadDxfTriangulations(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, 'resources', 'dxf_files')

        self.extension = "dxf"

        self.file_001_path = os.path.join(self.resource_dir, "file_001." + self.extension)
        self.file_002_path = os.path.join(self.resource_dir, "file_002." + self.extension)

    def test_file_001(self):
        models_collection = CadReader.readFile(self.file_001_path)
        self.assertEqual(len(models_collection), 2)

        n_triangulations = len([m.getTriangulations() for m in models_collection])
        self.assertEqual(n_triangulations, 2)

    def test_file_002(self):
        models_collection = CadReader.readFile(self.file_002_path)
        self.assertEqual(len(models_collection), 2)

        n_triangulations = sum([len(m.getTriangulations()) for m in models_collection])
        self.assertEqual(n_triangulations, 3)


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING, Union, List, Optional

import numpy as np
import shapely.geometry as shageo

from geogeometry.geometry.vector.Vector import Vector

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylineQuerier(object):

    def __init__(self, polyline: 'Polyline'):
        self.polyline = polyline

    def arePointsInside(self, test_nodes: Union[List, np.ndarray]) -> List[bool]:
        if self.polyline.getDimensions() != 2:
            raise ValueError("Inclusion queries only works on 2D.")

        poly = self.polyline.getNodes()
        if not self.polyline.isClosed():
            poly = np.array(list(poly) + [poly[0]])

        polygon = shageo.Polygon((poly))
        shageo_nodes = shageo.MultiPoint(np.array(test_nodes)[:, :2]).geoms

        results = []
        for n in shageo_nodes:
            results += [polygon.contains(n)]

        return results

    def isPointInside(self, test_node: np.ndarray) -> bool:
        results = self.arePointsInside(np.array([test_node]))
        return results[0]

    def getPointDistanceFromOrigin(self, point: np.ndarray) -> Optional[float]:
        """
        Point must be a point within the polyline. Typically coming from an intersection process.
        :param point: Numpy array 2D or 3D
        :return: Distance from the origin along the polyline.
        """
        distance = 0
        for i, n in enumerate(self.polyline.getNodes()[:-1]):
            segment = Vector(n0=n, n1=self.polyline.getNodes()[i+1])
            if self.checkNodeWithinSegment(point=point, segment=segment.getNodes()):
                distance += Vector(n0=n, n1=point).getLength()
                break
            else:
                distance += segment.getLength()
        else:
            raise ValueError("Point not along polyline.")

        return float(distance)

    def getPointAtDistanceFromOrigin(self, distance: float) -> Optional[np.ndarray]:

        if distance > self.polyline.getLength():
            raise ValueError("Distance is greater than polyline length.")
        elif distance == self.polyline.getLength():
            return self.polyline.getNodes()[-1]

        current_distance = 0
        for i, n in enumerate(self.polyline.getNodes()[:-1]):
            segment = Vector(n0=n, n1=self.polyline.getNodes()[i + 1])

            if current_distance + segment.getLength() > distance:
                out_point = n + segment.getNormalizedVector().getN1() * (distance - current_distance)
                return out_point
            elif current_distance + segment.getLength() < distance:
                current_distance += segment.getLength()
            else:
                return self.polyline.getNodes()[i + 1]

    @staticmethod
    def checkNodeWithinSegment(point: np.ndarray, segment: Union[List[float], np.ndarray],
                               cross_tolerance: float = 1e-6, dot_tolerance: float = 1e-12) -> bool:
        segment = np.array(segment)

        segment_vector = Vector(n0=segment[0], n1=segment[1])
        test_vector = Vector(n0=segment[0], n1=point)

        if test_vector.getLength() > segment_vector.getLength():
            return False

        cross_vector = Vector(n0=np.cross(segment_vector.getNormalizedVector(), test_vector.getNormalizedVector()))

        # Si colinean checkeo que node esta en segment
        if cross_vector.getLength() < cross_tolerance:
            # Debe ser mayor que 0 para que el punto este dentro del segment.
            dot = np.dot(segment_vector.getNormalizedVector(), test_vector.getNormalizedVector())
            if dot >= -dot_tolerance:  # - toleran ya que puede estar muy cerca por el otro lado
                return True

        return False



import os
import unittest
from geogeometry import CadReader


class TestReadRhinoPolylines(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, 'resources', 'rhino_files')

        self.extension = "3dm"

        self.file_003_path = os.path.join(self.resource_dir, "file_003." + self.extension)
        self.file_004_path = os.path.join(self.resource_dir, "file_004." + self.extension)
        self.file_005_path = os.path.join(self.resource_dir, "file_005." + self.extension)

    def test_file_003(self):
        """
        :case: Multiple 2-points 3D polylines (segments)
        """
        models_collection = CadReader.readFile(self.file_003_path)
        self.assertEqual(len(models_collection), 15)  # Layers

        n_polylines = len([m.getPolylines() for m in models_collection])
        self.assertEqual(n_polylines, 15)

    def test_file_004(self):
        """
        :case: Multiple 3D polylines
        """
        models_collection = CadReader.readFile(self.file_004_path)
        self.assertEqual(len(models_collection), 47)  # Layers

        n_polylines = sum([len(m.getPolylines()) for m in models_collection])
        self.assertEqual(n_polylines, 47)

    def test_file_005(self):
        """
        :case: Multiple 3D polylines
        """
        models_collection = CadReader.readFile(self.file_005_path)
        self.assertEqual(len(models_collection), 16)  # Layers

        n_polylines = sum([len(m.getPolylines()) for m in models_collection])
        self.assertEqual(n_polylines, 16)


if __name__ == "__main__":
    unittest.main()


from typing import Optional, Union, List

import numpy as np


class TriangulationProperties(object):

    def __init__(self):
        super().__init__()

        self.nodes: Optional[np.ndarray] = None
        self.faces: Optional[np.ndarray] = None

        self.markers: Optional[np.ndarray] = None

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        if isinstance(nodes, list):
            nodes = np.array(nodes)
        self.nodes = nodes

    def setFaces(self, faces: Union[List[int], np.ndarray[int]]) -> None:
        if isinstance(faces, list):
            faces = np.array(faces).astype(int)
        self.faces = faces

    def setMarkers(self, markers: Union[List[int], np.ndarray[int]]) -> None:
        if isinstance(markers, list):
            markers = np.array(markers).astype(int)
        self.markers = markers

    def getNodes(self) -> np.ndarray:
        return self.nodes

    def getFaces(self) -> np.ndarray:
        return self.faces

    def getMarkers(self) -> np.ndarray:
        return self.markers



from typing import Union, List, Optional

import numpy as np

from geogeometry.geometry.shared.BaseGeometry import BaseGeometry
from geogeometry.geometry.triangulation.components.TriangulationProperties import TriangulationProperties


class Triangulation(BaseGeometry, TriangulationProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    def translate(self, translation_vector: Union[List, np.ndarray]) -> None: ...
    def calculateLimits(self) -> None: ...
    def calculateCentroid(self) -> None: ...



from typing import Optional

import numpy as np
import rhino3dm

from geogeometry.geometry.model.GeometryModel import GeometryModel
from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection
from geogeometry.geometry.polyline.Polyline import Polyline

from geogeometry.geometry.triangulation.Triangulation import Triangulation


class RhinoReader(object):

    @staticmethod
    def readFile(filepath: str) -> Optional[GeometryModelsCollection]:
        rhino_model = rhino3dm.File3dm.Read(filepath)

        # print(dir(rhino_model))

        models_collection = GeometryModelsCollection()
        for layer in rhino_model.Layers:
            layer_model = GeometryModel(name=layer.Name)
            models_collection.addElement(layer_model)

        for obj in rhino_model.Objects:
            obj_type = obj.Geometry.ObjectType.name

            layer_index = obj.Attributes.LayerIndex
            layer_name = rhino_model.Layers[layer_index].Name

            if obj_type == 'Mesh':

                nodes = np.array([[n.X, n.Y, n.Z] for n in obj.Geometry.Vertices])
                faces = np.array(obj.Geometry.Faces)[:, :3]

                t = Triangulation()
                t.setNodes(nodes=nodes)
                t.setFaces(faces=faces)

                models_collection[layer_name].addTriangulation(t)
            elif obj_type == "Curve":
                if hasattr(obj.Geometry, "Line"):
                    start = obj.Geometry.Line.From
                    end = obj.Geometry.Line.To

                    n0 = np.array([start.X, start.Y, start.Z])
                    n1 = np.array([end.X, end.Y, end.Z])

                    nodes = [n0, n1]

                else:
                    polyline = obj.Geometry.TryGetPolyline()

                    nodes = []
                    for i in range(polyline.Count):
                        point = obj.Geometry.Point(i)
                        nodes += [[point.X, point.Y, point.Z]]

                p = Polyline()
                p.setNodes(nodes=nodes)
                models_collection[layer_name].addPolyline(p)

            else:
                raise ValueError(f"obj_type: {obj_type}")

        models_collection.deleteEmptyModels()

        return models_collection



from typing import Optional

from geogeometry.geometry.shared.BaseGeometryCollection import BaseGeometryCollection


class PolylinesCollection(BaseGeometryCollection):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)



import math
from typing import TYPE_CHECKING, List

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorTransformer(object):

    def __init__(self, vector: 'Vector'):
        self.vector: 'Vector' = vector

    def reverse(self) -> List[np.ndarray]:
        n0 = -1 * self.vector.getN0()
        n1 = -1 * self.vector.getN1()
        return [n0, n1]

    def rotate2d(self, angle: float) -> List[np.ndarray]:
        """
        ANTI-CLOCKWISE. Rotation axis at the origin (0,0).
        :param angle: In degrees.
        """
        angle = math.radians(angle)
        rot_matrix = np.array([
                                [math.cos(angle), -math.sin(angle)],
                                [math.sin(angle), math.cos(angle)],
                            ])

        n0 = rot_matrix.dot(self.vector.getN0())
        n1 = rot_matrix.dot(self.vector.getN1())

        return [n0, n1]



from typing import Optional

from geogeometry.geo_io.readers.RhinoReader import RhinoReader
from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection
from geogeometry.geo_io.readers.DxfReader import DxfReader


class CadReader(object):

    @staticmethod
    def readFile(filepath: str) -> Optional[GeometryModelsCollection]:
        extension = filepath.split('.')[-1].lower()

        if extension == "dxf":
            return DxfReader.readFile(filepath)
        elif extension == '3dm':
            return RhinoReader.readFile(filepath)
        else:
            raise ValueError(f"File extension '.{extension}' not supported.")



import copy
from typing import TYPE_CHECKING, Tuple

import numpy as np

from geogeometry.geometry.operations.Angles import Angles

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorMetrics(object):

    def __init__(self, vector: 'Vector'):
        self.vector: 'Vector' = vector

    def calculateMetrics(self) -> None:
        self.calculateLength()
        self.calculateNormalizedVector()
        self.calculateDipDipdir()

    def calculateLength(self) -> None:
        diff = self.vector.getN1() - self.vector.getN0()
        length = np.linalg.norm(diff)
        self.vector.setLength(length=length)

        self.vector.setUnitary(unitary=length == 1.)

    def calculateNormalizedVector(self) -> None:

        if self.vector.isUnitary():
            self.vector.setNormalizedVector(normalized_vector=copy.deepcopy(self.vector))
            return

        vector = np.array(self.vector.getN1() - self.vector.getN0())
        norm = np.linalg.norm(vector)

        if norm == 0.:
            raise ValueError("Origin as vector.")

        self.vector.createNormalizedVector(n=vector/norm)

    def calculateDipDipdir(self) -> None:

        if self.vector.getNormalizedVector() is None:
            self.calculateNormalizedVector()

        norm_vector = self.vector.getNormalizedVector()

        dipdir = None
        if norm_vector.getN0().shape[0] == 2:
            dip = 90.
        else:
            if abs(norm_vector.getN1()[2]) == 1.:
                dip, dipdir = 0, 0
            else:
                dip = Angles.calculateAngleBetweenVectorAndAxis(v=norm_vector, axis_id='z')
                if dip > 90.:
                    dip = Angles.calculateAngleBetweenVectorAndAxis(v=norm_vector.reverse(), axis_id='z')

        if dipdir is None:
            dipdir = Angles.calculateAngleFromThreePoints([0., 1.], [0., 0.], norm_vector.getN1()[:2])

        if norm_vector.getN1()[0] < 0.:
            dipdir = 360. - dipdir

        self.vector.setDip(dip=round(dip, 2))
        self.vector.setDipdir(dipdir=round(dipdir, 2))



from typing import Optional, Union

import ezdxf
import ezdxf.entities

import numpy as np

from geogeometry.geometry.triangulation.Triangulation import Triangulation
from geogeometry.geometry.polyline.Polyline import Polyline


class DxfEntityHandler(object):

    @staticmethod
    def handlePolyline(entity: ezdxf.entities.Polyline) -> Optional[Union[Triangulation, Polyline]]:

        if entity.is_poly_face_mesh:
            return DxfEntityHandler.handlePolyfaceMesh(entity=entity)
        else:
            return DxfEntityHandler.handleActualPolyline(entity=entity)

    @staticmethod
    def handlePolyfaceMesh(entity: ezdxf.entities.Polyline) -> Optional[Triangulation]:

        vertices = []
        faces = []
        for vertex in entity.vertices:
            if vertex.is_face_record:
                # Just considers triangulations
                indices = [vertex.get_dxf_attrib(name, 0) for name in ('vtx0', 'vtx1', 'vtx2')]
                # indices = [vertex.dxf.vtx0, vertex.dxf.vtx1, vertex.dxf.vtx2, vertex.dxf.vtx3]

                indices = [idx - 1 for idx in indices if idx != 0]
                faces += [indices]

            elif vertex.is_poly_face_mesh_vertex:
                x, y, z = vertex.dxf.location
                vertices += [[x, y, z]]

        vertices = np.array(vertices)

        t = Triangulation()
        t.setNodes(nodes=vertices)
        t.setFaces(faces=faces)

        return t

    @staticmethod
    def handleActualPolyline(entity: ezdxf.entities.Polyline) -> Optional[Polyline]:

        nodes = [list(p) for p in entity.points()]
        if entity.is_closed:
            nodes += [nodes[0]]

        p = Polyline()
        p.setNodes(nodes=nodes)
        return p

    @staticmethod
    def handleLine(entity: ezdxf.entities.Line) -> Optional[Polyline]:
        n0 = np.array(entity.dxf.start)
        n1 = np.array(entity.dxf.end)

        p = Polyline()
        p.setNodes(nodes=[n0, n1])
        return p



from typing import Optional, List

from geogeometry.geometry.model.GeometryModel import GeometryModel
from geogeometry.geometry.shared.BaseGeometryCollection import BaseGeometryCollection


class GeometryModelsCollection(BaseGeometryCollection):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    def deleteEmptyModels(self) -> None:

        empty_models: List[GeometryModel] = [e for e in self if not len(e)]
        for m in empty_models:
            self.deleteElement(identifier=m.getName())



from typing import Union, List, Optional

import numpy as np

from geogeometry.geometry.shared.BaseGeometry import BaseGeometry
from geogeometry.geometry.model.components.GeometryModelProperties import GeometryModelProperties


class GeometryModel(BaseGeometry, GeometryModelProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    def __len__(self) -> int:
        total = 0
        total += len(self.getTriangulations())
        total += len(self.getPolylines())
        return total

    def translate(self, translation_vector: Union[List, np.ndarray]) -> None: ...
    def calculateLimits(self) -> None: ...
    def calculateCentroid(self) -> None: ...



from typing import Optional

import numpy as np


class BaseGeometryProperties(object):

    def __init__(self, name: Optional[str] = None):
        super().__init__()

        self.name: Optional[str] = name
        self.color: str = 'red'
        self.opacity: float = 1.0

        self.id: int = id(self)

        self.limits: Optional[np.ndarray] = None
        self.centroid: Optional[np.ndarray] = None

    def setName(self, name: str) -> None:
        self.name = name

    def setColor(self, color: str) -> None:
        self.color = color

    def setOpacity(self, opacity: float) -> None:
        self.opacity = opacity

    def setLimits(self, limits: np.ndarray) -> None:
        self.limits = limits

    def setCentroid(self, centroid: np.ndarray) -> None:
        self.centroid = centroid

    def getName(self) -> str:
        return self.name

    def getColor(self) -> str:
        return self.color

    def getOpacity(self) -> float:
        return self.opacity

    def getLimits(self) -> Optional[np.ndarray]:
        return self.limits

    def getCentroid(self) -> Optional[np.ndarray]:
        return self.centroid



from typing import Optional, Literal, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylineMetrics(object):

    def __init__(self, polyline: 'Polyline'):
        self.polyline: 'Polyline' = polyline

    def calculateMetrics(self) -> None:
        nodes = self.polyline.getNodes()

        self.polyline.setDimensions(dimensions=nodes.shape[1])

        closed = False
        if not np.linalg.norm(nodes[0] - nodes[-1]):
            closed = True
        self.polyline.setClosed(closed=closed)

        self.calculateLimits()
        self.calculateLength()
        self.calculateMidPoint()

        if self.polyline.getDimensions() == 2 and self.polyline.isClosed():
            self.calculate2DArea()

    def calculateLength(self):
        nodes = self.polyline.getNodes()
        ds = [np.linalg.norm(n - nodes[i + 1]) for i, n in enumerate(nodes[:-1])]
        self.polyline.setLength(length=np.sum(ds))

    def calculateLimits(self) -> None:
        mins = np.min(self.polyline.getNodes(), axis=0)
        maxs = np.max(self.polyline.getNodes(), axis=0)

        self.polyline.setLimits(np.array([mins, maxs]))

    def calculate2DArea(self) -> None:
        """
        Considers XY space, even if nodes dimensions is greater than 2.
        Gets area even if the polyline is open.
        """
        nodes = self.polyline.getNodes()
        if self.polyline.isClosed():
            x, y = nodes[:-1][:, 0], nodes[:-1][:, 1]
        else:
            x, y = nodes[:, 0], nodes[:, 1]

        area = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

        self.polyline.setArea(area=area)

    def calculateMidPoint(self) -> None:
        """
        Mid at (total length) / 2
        """
        mid_point = self.polyline.getPointAtDistanceFromOrigin(distance=self.polyline.getLength() / 2.)
        self.polyline.setMidPoint(mid_point)


from typing import List, TypeVar, Iterator, Optional, Union

T = TypeVar('T')


class BaseGeometryCollection(object):

    def __init__(self, name: Optional[str] = None):
        self.name: Optional[str] = name
        self.elements: List[T] = []

    def __len__(self) -> int:
        return len(self.elements)

    def __iter__(self) -> Iterator[T]:
        for e in self.elements:
            yield e

    def __getitem__(self, identifier: Union[int, str]) -> T:
        if isinstance(identifier, int):
            return self.elements[identifier]
        else:
            for e in self.elements:
                if e.getName() == identifier:
                    return e
            else:
                raise ValueError(f"Element '{identifier}' not found in collection.")

    def addElement(self, element: T) -> None:
        self.elements += [element]

    def deleteElement(self, identifier: Union[int, str]):
        for e in self.elements:
            if e.getName() == identifier:
                self.elements.remove(e)
                break
        else:
            raise ValueError(f"Element '{identifier}' not found in collection.")



from typing import Optional, TYPE_CHECKING

from geogeometry.geometry.polyline.PolylinesCollection import PolylinesCollection
from geogeometry.geometry.triangulation.TriangulationsCollection import TriangulationsCollection

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline
    from geogeometry.geometry.triangulation.Triangulation import Triangulation


class GeometryModelProperties(object):

    def __init__(self):
        super().__init__()
        self.triangulations: Optional[TriangulationsCollection] = TriangulationsCollection()
        self.polylines: Optional[PolylinesCollection] = PolylinesCollection()

    def getTriangulations(self) -> Optional[TriangulationsCollection]:
        return self.triangulations

    def getPolylines(self) -> Optional[PolylinesCollection]:
        return self.polylines

    def addTriangulation(self, triangulation: 'Triangulation') -> None:
        self.triangulations.addElement(triangulation)

    def addPolyline(self, polyline: 'Polyline') -> None:
        self.polylines.addElement(polyline)



from typing import Union, List, Optional

import numpy as np

from geogeometry.geometry.shared.BaseGeometry import BaseGeometry
from geogeometry.geometry.polyline.components.PolylineProperties import PolylineProperties

from geogeometry.geometry.polyline.components.PolylineMetrics import PolylineMetrics
from geogeometry.geometry.polyline.components.PolylineQuerier import PolylineQuerier



class Polyline(BaseGeometry, PolylineProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.metrics: PolylineMetrics = PolylineMetrics(self)
        self.querier: PolylineQuerier = PolylineQuerier(self)

    def translate(self, translation_vector: Union[List, np.ndarray]) -> None: ...

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        super().setNodes(nodes=nodes)
        self.metrics.calculateMetrics()

    def calculate2DArea(self) -> None:
        self.metrics.calculate2DArea()

    def calculateLimits(self) -> None:
        self.metrics.calculateLimits()

    def calculateCentroid(self) -> None: ...

    def getPointAtDistanceFromOrigin(self, distance: float) -> Optional[np.ndarray]:
        return self.querier.getPointAtDistanceFromOrigin(distance=distance)


import unittest
from geogeometry import Triangulation


class TestBaseGeometry(unittest.TestCase):

    def setUp(self):
        pass

    def test_creation(self):
        t = Triangulation()
        self.assertIsNotNone(t)
        self.assertIsInstance(t, Triangulation)


if __name__ == "__main__":
    unittest.main()



import math
from typing import Optional, Tuple, Any, Callable, List

import numpy as np

from geogeometry.geometry.vector.components.VectorMetrics import VectorMetrics
from geogeometry.geometry.vector.components.VectorProperties import VectorProperties
from geogeometry.geometry.vector.components.VectorTransformer import VectorTransformer


def nodesToVector(func: Callable[..., List[np.ndarray]]) -> Callable[..., 'Vector']:
    def inner(vector: 'Vector', *args, **kwargs) -> 'Vector':
        nodes = func(vector, *args, **kwargs)
        return Vector(n0=nodes[0], n1=nodes[1])

    return inner


class Vector(VectorProperties):

    def __init__(self, n0: np.ndarray, n1: Optional[np.ndarray] = None):
        super().__init__()

        if n1 is None:
            self.setN0(n0=np.zeros(n0.shape[0]))
            self.setN1(n1=n0)
        else:
            self.setN0(n0=n0)
            self.setN1(n1=n1)

        self.metrics: VectorMetrics = VectorMetrics(self)
        self.transformer: VectorTransformer = VectorTransformer(self)

        self.metrics.calculateMetrics()

    def createNormalizedVector(self, n: np.ndarray) -> None:
        self.setNormalizedVector(normalized_vector=Vector(n0=n))

    @nodesToVector
    def reverse(self) -> List[np.ndarray]:
        return self.transformer.reverse()

    @nodesToVector
    def rotate2d(self, angle: float) -> List[np.ndarray]:
        return self.transformer.rotate2d(angle=angle)



from typing import Optional, Union, List, Literal

import numpy as np


class PolylineProperties(object):

    def __init__(self):

        self.nodes: Optional[np.ndarray] = None
        self.segments: Optional[np.ndarray] = None

        self.dimensions: Optional[Literal[2, 3]] = None
        self.closed: Optional[bool] = None

        self.length: Optional[float] = None

        self.mid_point: Optional[np.ndarray] = None

        # 2D properties
        self.area: Optional[float] = None

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        self.nodes = np.array(nodes)
        self.segments = [[j, j + 1] for j, n in enumerate(self.nodes[:-1])]

    def setDimensions(self, dimensions: Literal[2, 3]) -> None:
        self.dimensions = dimensions

    def setClosed(self, closed: bool) -> None:
        self.closed = closed

    def setLength(self, length: float) -> None:
        self.length = length

    def setMidPoint(self, mid_point: np.ndarray) -> None:
        self.mid_point = mid_point

    def setArea(self, area: float) -> None:
        self.area = area

    def getNodes(self) -> Optional[np.ndarray]:
        return self.nodes

    def getSegments(self) -> Optional[np.ndarray]:
        return self.segments

    def getDimensions(self) -> Optional[int]:
        return self.dimensions

    def isClosed(self) -> Optional[bool]:
        return self.closed

    def getLength(self) -> Optional[float]:
        return self.length

    def getMidPoints(self) -> Optional[np.ndarray]:
        return self.mid_point

    def getArea(self) -> Optional[float]:
        return self.area



from abc import ABC, abstractmethod
from typing import Union, List, TypeVar, Optional

import gzip
import lzma
import pickle

import numpy as np

from geogeometry.geometry.shared.BaseGeometryProperties import BaseGeometryProperties


T = TypeVar('T')


class BaseGeometry(ABC, BaseGeometryProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    @abstractmethod
    def translate(self, translation_vector: Union[List, np.ndarray]) -> None:
        pass

    @abstractmethod
    def calculateLimits(self) -> None:
        pass

    @abstractmethod
    def calculateCentroid(self) -> None:
        pass

    # IO
    def save(self, savepath: str) -> None:
        if '.pkl.gz' in savepath:
            pickle.dump(self, gzip.open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)
        elif 'pkl.xz' in savepath:
            pickle.dump(self, lzma.open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)
        else:
            pickle.dump(self, open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)

        print(f'{self.__class__.__name__} instance saved at: "' + savepath + '"')

    @staticmethod
    def load(load_path) -> T:
        if '.pkl.gz' in load_path:
            return pickle.load(gzip.open(load_path, 'rb'))
        elif 'pkl.xz' in load_path:
            return pickle.load(lzma.open(load_path, 'rb'))
        else:
            return pickle.load(open(load_path, 'rb'))
